# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import itertools

def increment_coord(c, shape):
    i = len(shape)-1
    done = False
    while not done:
        c[i] += 1
        if c[i]>=shape[i]:
            c[i] = 0
            i -= 1
            if i<0: raise IndexError
        else:
            done = True

# ndindex is called a lot, so cache results (shape parameter must be a tuple so it can go in a dict)
ndindex_cache = {}

def ndindex(shape):
    global ndindex_cache
    try:
        return ndindex_cache[shape]
    except:
        pass
    coords = []
    c = [0]*len(shape)
    try:
        while 1:
            coords.append(tuple(c))
            increment_coord(c, shape)
    except IndexError:
        pass
    ndindex_cache[shape] = coords
    return coords

class Hypergrid(object):
    """
    A multi-dimensional array that supports a subset of the numpy API, used when the numpy module
    is unavailable (e.g. on the iPhone)
    """
    def __init__(self, shape, factory=int):
        self.shape = shape
        self.factory = factory
        self._grid = {}
        # unlike numpy arrays, values are stored in a dict, but still indexed by tuple keys
        for c in ndindex(self.shape):
            self._grid[c] = self.factory()
        
    def __setitem__(self, key, value):
        self._grid[key] = value
        
    def __getitem__(self, key):
        # fixme: slices
        try:
            return self._grid[key]
        except:
            pass
        # check for slices, unlike real numpy the returned array will not be a view of self
        try:
            if any(isinstance(x, slice) for x in key):
                ranges = []
                for i,x in enumerate(key):
                    # if slice, use start and stop, treat int as slice of length 1
                    ranges.append((max(0,x.start), min(self.shape[i],x.stop)) if isinstance(x,slice) else (x,x+1))
                if all(s[1]-s[0]>0 for s in ranges):
                    subgrid = Hypergrid(tuple(s[1]-s[0] for s in ranges))
                    # fill subgrid by offsetting according to slices
                    for c in ndindex(subgrid.shape):
                        pc = tuple(i+s[0] for i,s in itertools.izip(c, ranges))
                        subgrid[c] = self._grid[pc]
                    return subgrid
        except:
            raise
        raise KeyError(key)
        
    def __contains__(self, item):
        return item in self._grid.values()
        
    def _flat(self):
        for c in ndindex(self.shape):
            yield self._grid[c]
    flat = property(_flat, None)
    
    ndim = property(lambda self: len(self.shape), None)
    
    # size property is the total number of entries in the "array"
    size = property(lambda self: len(self._grid), None)
    
    def reshape(self, newshape):
        newgrid = {}
        for newc,oldc in itertools.izip(ndindex(newshape), ndindex(self.shape)):
            newgrid[newc] = self._grid[oldc]
        self.shape = newshape
        self._grid = newgrid
        return self
        
    def copy(self):
        newgrid = type(self)(self.shape, self.factory)
        newgrid._grid = self._grid.copy()
        return newgrid
        
        
def empty(shape, factory=int):
    return Hypergrid(shape, factory)
    
def zeros(shape, factory=int):
    return Hypergrid(shape, factory)
    
def ndenumerate(array):
    for c in array._grid:
        yield (c, array._grid[c])
        
def array(items):
    a = Hypergrid((len(items),))
    for i,x in enumerate(items):
        a[i,] = x
    return a
    
    
    
def test():
    a = array(range(12))
    assert a[0,]==0
    assert a[11,]==11
    a.reshape((3,4))
    assert a.shape==(3,4)
    assert a.ndim==2
    assert a.size==12
    assert a[1,0] == 4
    
    b = a[1:5,0:2]
    assert b.shape==(2,2)
    assert all(b[x,y]==a[x+1,y] for x,y in ndindex(b.shape))
    
    return a
    
    
if __name__=='__main__':
    test()
    print 'Tests passed'
    